/*

  G N O K I I

  A Linux/Unix toolset and driver for Nokia mobile phones.

  Copyright (C) 1999, 2000 Hugh Blemings & Pavel Jank ml.

  Released under the terms of the GNU GPL, see file COPYING for more details.
	
  Provides a generic API for accessing functions on the phone, wherever
  possible hiding the model specific details.

  The underlying code should run in it's own thread to allow communications to
  the phone to be run independantly of mailing code that calls these API
  functions.

  Unless otherwise noted, all functions herein block until they complete.  The
  functions themselves are defined in a structure in gsm-common.h.

  Last modification: Mon Mar 20 22:02:15 CET 2000
  Modified by Pavel Jank ml. <Pavel.Janik@linux.cz>

*/

#include <stdio.h>
#include <string.h>

#include "misc.h"
#include "gsm-common.h"
#include "rlp-common.h"
#include "fbus-3810.h"
#include "fbus-6110.h"
#include "mbus-2110.h"
#include "mbus-6160.h"
#include "mbus-640.h"

/* GSM_LinkOK is set to true once normal communications with the phone have
   been established. */

bool *GSM_LinkOK;

/* Define pointer to the GSM_Functions structure used by external code to call
   relevant API functions. This structure is defined in gsm-common.h. */

GSM_Functions *GSM;

/* Define pointer to the GSM_Information structure used by external code to
   obtain information that varies from model to model. This structure is also
   defined in gsm-common.h */

GSM_Information		*GSM_Info;

/* Initialise interface to the phone. Model number should be a string such as
   3810, 5110, 6110 etc. Device is the serial port to use e.g. /dev/ttyS0, the
   user must have write permission to the device. */


GSM_Error GSM_Initialise(char *model, char *device, char *initlength, GSM_ConnectionType connection, void (*rlp_callback)(RLP_F96Frame *frame))
{
  bool found_match=false;

  /* Scan through the models supported by the FB38 code and see if we get a
     match. */

  if (strstr(FB38_Information.Models, model) != NULL) {
    found_match = true;

    /* Set pointers to relevant FB38 addresses */

    GSM = &FB38_Functions;
    GSM_Info = &FB38_Information;
    GSM_LinkOK = &FB38_LinkOK;
  }
  else

  /* Scan through the models supported by the FB61 code and see if we get a
     match. */

    if (strstr(FB61_Information.Models, model) != NULL) {
      found_match = true;

      /* Set pointers to relevant FB61 addresses */

      GSM = &FB61_Functions;
      GSM_Info = &FB61_Information;
      GSM_LinkOK = &FB61_LinkOK;
    }
#ifndef WIN32	/* MB21 not supported in win32 */
    else 
	/* Scan through models supported by the MB21 code... */
    if (strstr(MB21_Information.Models, model) != NULL) {
      found_match = true;

      /* Set pointers to relevant MB21 addresses */

      GSM = &MB21_Functions;
      GSM_Info = &MB21_Information;
      GSM_LinkOK = &MB21_LinkOK;
    }
    else 
	/* Scan through models supported by the MB6160 code... */
    if (strstr(MB61_Information.Models, model) != NULL) {
      found_match = true;

      /* Set pointers to relevant MB61 addresses */

      GSM = &MB61_Functions;
      GSM_Info = &MB61_Information;
      GSM_LinkOK = &MB61_LinkOK;
    }
    else 
	/* Scan through models supported by the MB640 code... */
    if (strstr(MB640_Information.Models, model) != NULL) {
      found_match = true;

      /* Set pointers to relevant MB640 addresses */

      GSM = &MB640_Functions;
      GSM_Info = &MB640_Information;
      GSM_LinkOK = &MB640_LinkOK;
    }
#endif /* WIN32 */ 

  /* If we didn't get a model match, return error code. */

  if (found_match == false)
    return (GE_UNKNOWNMODEL);

  /* Now call model specific initialisation code. */

  return (GSM->Initialise(device, initlength, connection, rlp_callback));
}
